


#define _CRT_SECURE_NO_WARNINGS
#define _CRT_NONSTDC_NO_WARNINGS

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "ad_globl.h"
#include "ad_prime.h"
#include "iad_type.h"
#include "iad_pub.h"
#include "iad_io.h"
#include "iad_calc.h"
#include "iad_util.h"
#include "mygetopt.h"
#include "version.h"
#include "mc_lost.h"

extern char *optarg;
extern int optind;





static void
print_version (void)
{
  fprintf (stderr, "iad %s\n", Version);
  fprintf (stderr, "Copyright 2007 Scott Prahl, prahl@bme.ogi.edu\n");
  fprintf (stderr, "          (see Applied Optics, 32:559-568, 1993)\n");
  exit (0);
}



static void
print_usage (void)
{
  fprintf (stderr, "iad %s\n\n", Version);
  fprintf (stderr, "iad finds optical properties from measurements\n\n");
  fprintf (stderr, "Usage:  iad [options] input\n\n");
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "  -1 '# # # # #'   reflection sphere parameters \n");
  fprintf (stderr, "  -2 '# # # # #'   transmission sphere parameters \n");
  fprintf (stderr, "  -a #             use this albedo \n");
  fprintf (stderr, "  -b #             use this optical thickness \n");
  fprintf (stderr, "  -e #             error tolerance (default 0.0001) \n");
  fprintf (stderr, "  -g #             default anisotropy (default 0) \n");
  fprintf (stderr, "  -h               display help\n");
  fprintf (stderr, "  -m               machine readable output\n");
  fprintf (stderr, "  -M #             number of Monte Carlo iterations\n");
  fprintf (stderr,
	   "  -n #             specify index of refraction of slab\n");
  fprintf (stderr,
	   "  -N #             specify index of refraction of slides\n");
  fprintf (stderr,
	   "  -o filename      explicitly specify filename for output\n");
  fprintf (stderr,
	   "  -p #             # of Monte Carlo photons (default 100000)\n");
  fprintf (stderr,
	   "                   a negative number is max MC time in milliseconds\n");
  fprintf (stderr,
	   "  -q #             number of quadrature points (default=8)\n");
  fprintf (stderr,
	   "  -Q               quiet -- silence output to stderr\n\n");
  fprintf (stderr, "  -r #             total reflection measurement\n");
  fprintf (stderr,
	   "  -s               show sphere and MC light loss effects\n");
  fprintf (stderr, "  -S #             number of spheres used\n");
  fprintf (stderr, "  -t #             total transmission measurement\n");
  fprintf (stderr,
	   "  -u #             unscattered transmission measurement\n");
  fprintf (stderr, "  -v               version information\n");
  fprintf (stderr, "Examples:\n");
  fprintf (stderr,
	   "  iad data                    optical values put in data.txt\n");
  fprintf (stderr,
	   "  iad -e 0.0001 data          better convergence to R & T values\n");
  fprintf (stderr,
	   "  iad -m data                 data.abg in machine readable format\n");
  fprintf (stderr,
	   "  iad -o out data             calculated values in out\n");
  fprintf (stderr,
	   "  iad -r 0.3                  R_total=0.3, b=inf, find albedo\n");
  fprintf (stderr,
	   "  iad -r 0.3 -t 0.4           R_total=0.3, T_total=0.4, find a,b,g\n");
  fprintf (stderr,
	   "  iad -r 0.3 -t 0.4 -n 1.5    R_total=0.3, T_total=0.4, n=1.5, find a,b\n");
  fprintf (stderr,
	   "  iad -r 0.3 -t 0.4           R_total=0.3, T_total=0.4, find a,b\n");
  fprintf (stderr, "  iad -p 1000 data            only 1000 photons\n");
  fprintf (stderr,
	   "  iad -p -100 data            allow only 100ms per iteration\n");
  fprintf (stderr, "  iad -q 4 data               four quadrature points\n");
  fprintf (stderr, "  iad -M 0 data               no MC    (iad)\n");
  fprintf (stderr,
	   "  iad -M 1 data               MC once  (iad -> MC -> iad)\n");
  fprintf (stderr,
	   "  iad -M 2 data               MC twice (iad -> MC -> iad -> MC -> iad)\n");
  fprintf (stderr,
	   "  iad -M 0 -q 4 data          Fast and crude conversion\n");
  fprintf (stderr,
	   "  iad -s data                 show sphere and MC light loss effects\n");
  fprintf (stderr,
	   "  apply iad data1 data2       process multiple files\n\n");
  fprintf (stderr, "Report bugs to <prahl@bme.ogi.edu>\n\n");
  exit (0);
}



static char *
strdup_together (char *s, char *t)
{
  char *both;

  if (s == NULL)
    {
      if (t == NULL)
	return NULL;
      return strdup (t);
    }

  if (t == NULL)
    return strdup (s);

  both = malloc (strlen (s) + strlen (t) + 1);
  if (both == NULL)
    fprintf (stderr, "Could not allocate memory for both strings.\n");

  strcpy (both, s);
  strcat (both, t);
  return both;
}




static double
seconds_elapsed (clock_t start_time)
{
  clock_t finish_time = clock ();
  return (double) (finish_time - start_time) / CLOCKS_PER_SEC;
}



static void
print_error_legend (void)
{
  fprintf (stderr,
	   "----------------- Sorry, but ... mistakes were made ---------------\n");
  fprintf (stderr, "  0-9 ==> Monte Carlo Iteration\n");
  fprintf (stderr, "   *  ==> Successful Calculation\n");
  fprintf (stderr,
	   "-------------------------------------------------------------------\n");
  fprintf (stderr, "   R  ==> Bad Reflection\n");
  fprintf (stderr, "   T  ==> Bad Transmission\n");
  fprintf (stderr, "   U  ==> Bad Unscattered Transmission\n");
  fprintf (stderr,
	   "   L  ==> Reflection + Transmission +  Monte Carlo Light Losses > 1\n");
  fprintf (stderr, "   !  ==> Reflection + Transmission + Unscattered > 1\n");
  fprintf (stderr, "   +  ==> Too many adding-doubling iterations\n\n");
}





static void
print_dot (clock_t start_time, int err, int count, int points,
	   int final, int *any_error)
{
  static int counter = 1;

  if (err == IAD_NO_ERROR)
    {
      if (final == 99)
	fprintf (stderr, "*");
      else
	fprintf (stderr, "%d", (int) fmod (final, 10));
    }
  else
    {
      *any_error = 1;
      if (err == IAD_TOO_MANY_ITERATIONS)
	fprintf (stderr, "+");

      else if (err == IAD_R_LT_ZERO ||
	       err == IAD_R_GT_ONE ||
	       err == IAD_RT_GT_ONE ||
	       err == IAD_RT_LT_MINIMUM || err == IAD_RD_LT_ZERO)
	fprintf (stderr, "R");

      else if (err == IAD_T_LT_ZERO ||
	       err == IAD_T_GT_ONE || err == IAD_RD_IS_ZERO_BUT_NOT_TD)
	fprintf (stderr, "T");

      else if (err == IAD_TU_LT_ZERO || err == IAD_TU_GT_ONE)
	fprintf (stderr, "U");

      else if (err == IAD_R_PLUS_T_GT_ONE ||
	       err == IAD_RT_PLUS_TT_GT_ONE ||
	       err == IAD_R_PLUS_T_PLUS_TU_GT_ONE)
	fprintf (stderr, "!");

      else if (err == IAD_EXCESSIVE_LIGHT_LOSS)
	fprintf (stderr, "L");

      else
	{
	  if (err >= IAD_AS_NOT_VALID && err <= IAD_TOO_MANY_LAYERS)
	    fprintf (stderr, "\nBad sphere parameter (error=%d)\n", err);
	  else if (err == IAD_MEMORY_ERROR)
	    fprintf (stderr, "\nMemory error\n");
	  else if (err == IAD_FILE_ERROR)
	    fprintf (stderr, "\nFile error\n");
	  exit (1);
	}
    }

  if (counter % 50 == 0)
    {
      double rate = (seconds_elapsed (start_time) / points);
      fprintf (stderr, "  %3d done (%5.2f s/pt)\n", points, rate);
    }
  else if (counter % 10 == 0)
    fprintf (stderr, " ");

  counter++;
  fflush (stderr);
}


static void
calculate_coefficients (struct measure_type m,
			struct invert_type r,
			double *LR, double *LT, double *musp, double *mua)
{
  double delta;
  *LR = 0;
  *LT = 0;
  *musp = 0;
  *mua = 0;

  if (r.error != IAD_NO_ERROR)
    return;

  Calculate_Distance (LR, LT, &delta);

  if (r.default_b == HUGE_VAL || r.b == HUGE_VAL)
    {
      double mus;
      mus = 1;
      *musp = mus * (1.0 - r.g);
      if (r.a != 0)
	*mua = mus * (1.0 - r.a) / r.a;
      else
	*mua = 0.0;
    }
  else
    {
      *musp = r.a * r.b / m.slab_thickness * (1.0 - r.g);
      *mua = (1 - r.a) * r.b / m.slab_thickness;
    }
}




static int
parse_string_into_array (char *s, double *a, int n)
{
  char *t, *last, *r;
  int i = 0;
  t = s;
  last = s + strlen (s);

  while (t < last)
    {


      r = t;
      while (*r != ' ' && *r != '\0')
	r++;
      *r = '\0';


      if (sscanf (t, "%lf", &(a[i])) == 0)
	return 1;
      i++;


      if (i == n)
	return 0;


      t = r + 1;
    }
  return 1;
}



int
main (int argc, char **argv)
{

  struct measure_type m;
  struct invert_type r;
  char *g_out_name = NULL;
  char c;
  int machine_readable_output = 0;
  int first_line = 1;
  int quiet = 0;
  int show_all = 0;

  long n_photons = 100000;
  int cl_quadrature_points = 8;
  int MC_iterations = 20;
  int any_error = 0;
  int process_command_line = 0;
  int params = 0;

  double cl_default_a = UNINITIALIZED;
  double cl_default_g = UNINITIALIZED;
  double cl_default_b = UNINITIALIZED;
  double cl_tolerance = UNINITIALIZED;

  double cl_sample_d = UNINITIALIZED;
  double cl_sample_n = UNINITIALIZED;
  double cl_slide_d = UNINITIALIZED;
  double cl_slide_n = UNINITIALIZED;

  double cl_UR1 = UNINITIALIZED;
  double cl_UT1 = UNINITIALIZED;
  double cl_Tc = UNINITIALIZED;

  double cl_num_spheres = UNINITIALIZED;
  double cl_sphere_one[5] = { UNINITIALIZED, UNINITIALIZED, UNINITIALIZED,
    UNINITIALIZED, UNINITIALIZED
  };
  double cl_sphere_two[5] = { UNINITIALIZED, UNINITIALIZED, UNINITIALIZED,
    UNINITIALIZED, UNINITIALIZED
  };

  clock_t start_time = clock ();




  while ((c =
	  my_getopt (argc, argv,
		     "?1:2:a:b:B:d:D:e:g:hn:N:mM:o:p:q:Qr:sS:t:u:v")) != EOF)
    {
      switch (c)
	{

	case '1':
	  parse_string_into_array (optarg, cl_sphere_one, 5);
	  break;

	case '2':
	  parse_string_into_array (optarg, cl_sphere_two, 5);
	  break;

	case 'a':
	  cl_default_a = strtod (optarg, NULL);
	  break;

	case 'b':
	  cl_default_b = strtod (optarg, NULL);
	  break;

	case 'B':
	  Set_Grid_Debugging ((unsigned int) strtod (optarg, NULL));
	  break;

	case 'd':
	  cl_sample_d = strtod (optarg, NULL);
	  break;

	case 'D':
	  cl_slide_d = strtod (optarg, NULL);
	  break;

	case 'e':
	  cl_tolerance = strtod (optarg, NULL);
	  break;

	case 'g':
	  cl_default_g = strtod (optarg, NULL);
	  break;

	case 'o':
	  g_out_name = strdup (optarg);
	  break;

	case 'n':
	  cl_sample_n = strtod (optarg, NULL);
	  break;

	case 'N':
	  cl_slide_n = strtod (optarg, NULL);
	  break;

	case 'p':
	  n_photons = (int) strtod (optarg, NULL);
	  break;

	case 'm':
	  machine_readable_output = 1;
	  quiet = 1;
	  break;

	case 'q':
	  cl_quadrature_points = (int) strtod (optarg, NULL);
	  if (cl_quadrature_points % 4 != 0)
	    {
	      fprintf (stderr,
		       "Number of quadrature points must be a multiple of 4\n");
	      exit (1);
	    }
	  break;

	case 'M':
	  MC_iterations = (int) strtod (optarg, NULL);
	  break;

	case 'Q':
	  quiet = 1;
	  break;

	case 'r':
	  cl_UR1 = strtod (optarg, NULL);
	  process_command_line = 1;
	  break;

	case 's':
	  show_all = 1;
	  break;

	case 'S':
	  cl_num_spheres = (int) strtod (optarg, NULL);
	  break;

	case 't':
	  cl_UT1 = strtod (optarg, NULL);
	  process_command_line = 1;
	  break;

	case 'u':
	  cl_Tc = strtod (optarg, NULL);
	  process_command_line = 1;
	  break;

	case 'v':
	  print_version ();
	  break;

	default:
	case 'h':
	case '?':
	  print_usage ();
	  break;
	}
    }

  argc -= optind;
  argv += optind;



  Initialize_Measure (&m);

  if (process_command_line)
    {
      r.method.quad_pts = 8;


      if (cl_sample_n != UNINITIALIZED)
	m.slab_index = cl_sample_n;

      if (cl_slide_n != UNINITIALIZED)
	{
	  m.slab_bottom_slide_index = cl_slide_n;
	  m.slab_top_slide_index = cl_slide_n;
	}

      if (cl_sample_d != UNINITIALIZED)
	m.slab_thickness = cl_sample_d;

      if (cl_slide_d != UNINITIALIZED)
	{
	  m.slab_bottom_slide_thickness = cl_slide_d;
	  m.slab_top_slide_thickness = cl_slide_d;
	}

      if (cl_UR1 != UNINITIALIZED)
	m.m_r = cl_UR1;

      if (cl_UT1 != UNINITIALIZED)
	m.m_t = cl_UT1;

      if (cl_Tc != UNINITIALIZED)
	m.m_u = cl_Tc;

      if (cl_num_spheres != UNINITIALIZED)
	m.num_spheres = (int) cl_num_spheres;

      if (cl_sphere_one[4] != UNINITIALIZED)
	{
	  double d_sample_r, d_entrance_r, d_detector_r;

	  m.d_sphere_r = cl_sphere_one[0];
	  d_sample_r = cl_sphere_one[1];
	  d_entrance_r = cl_sphere_one[2];
	  d_detector_r = cl_sphere_one[3];
	  m.rw_r = cl_sphere_one[4];

	  m.as_r = (d_sample_r / m.d_sphere_r) * (d_sample_r / m.d_sphere_r);
	  m.ae_r =
	    (d_entrance_r / m.d_sphere_r) * (d_entrance_r / m.d_sphere_r);
	  m.ad_r =
	    (d_detector_r / m.d_sphere_r) * (d_detector_r / m.d_sphere_r);
	  m.aw_r = 1.0 - m.as_r - m.ae_r - m.ad_r;

	  m.d_sphere_t = m.d_sphere_r;
	  m.as_t = m.as_r;
	  m.ae_t = m.ae_r;
	  m.ad_t = m.ad_r;
	  m.aw_t = m.aw_r;
	  m.rw_t = m.rw_r;
	}

      if (cl_sphere_two[4] != UNINITIALIZED)
	{
	  double d_sample_t, d_entrance_t, d_detector_t;

	  m.d_sphere_t = cl_sphere_two[0];
	  d_sample_t = cl_sphere_two[1];
	  d_entrance_t = cl_sphere_two[2];
	  d_detector_t = cl_sphere_two[3];
	  m.rw_t = cl_sphere_two[4];

	  m.as_t = (d_sample_t / m.d_sphere_t) * (d_sample_t / m.d_sphere_t);
	  m.ae_t =
	    (d_entrance_t / m.d_sphere_t) * (d_entrance_t / m.d_sphere_t);
	  m.ad_t =
	    (d_detector_t / m.d_sphere_t) * (d_detector_t / m.d_sphere_t);
	  m.aw_t = 1.0 - m.as_t - m.ae_t - m.ad_t;
	}

      m.num_measures = 3;
      if (m.m_t == 0)
	m.num_measures--;
      if (m.m_u == 0)
	m.num_measures--;
      params = m.num_measures;

      if (m.num_measures == 3)
	{

	  struct AD_slab_type s;
	  s.n_slab = m.slab_index;
	  s.n_top_slide = m.slab_top_slide_index;
	  s.n_bottom_slide = m.slab_bottom_slide_index;
	  s.b_top_slide = 0;
	  s.b_bottom_slide = 0;
	  cl_default_b = What_Is_B (s, m.m_u);
	}

      if (m.num_measures == 1)
	{
	  cl_default_b = HUGE_VAL;
	}


      if (m.num_spheres == 0)
	MC_iterations = 0;

      {
	static int rt_data_count = 0;
	static int mc_iter_count = 0;
	struct measure_type m_mc, m_none;
	struct invert_type r_mc, r_none;
	int mc_iter = 0;

	double ur1 = 0;
	double ut1 = 0;
	double uru = 0;
	double utu = 0;
	double ur1_lost = 0;
	double ut1_lost = 0;
	double uru_lost = 0;
	double utu_lost = 0;
	double mu_a_none = 0;
	double mu_a_sphere = 0;
	double mu_a_both = 0;
	double mu_a_mc = 0;
	double mu_sp_none = 0;
	double mu_sp_sphere = 0;
	double mu_sp_both = 0;
	double mu_sp_mc = 0;
	double LR = 0;
	double LT = 0;

	rt_data_count++;

	Initialize_Result (m, &r);
	r.method.quad_pts = cl_quadrature_points;

	if (cl_default_a != UNINITIALIZED)
	  {
	    r.default_a = cl_default_a;
	  }

	if (cl_default_b != UNINITIALIZED)
	  {
	    r.default_b = cl_default_b;
	  }

	if (cl_default_g != UNINITIALIZED)
	  {
	    r.default_g = cl_default_g;
	  }

	if (cl_tolerance != UNINITIALIZED)
	  {
	    r.tolerance = cl_tolerance;
	    r.MC_tolerance = cl_tolerance;
	  }
	r.search = determine_search (m, r);

	if (rt_data_count == 1 && !machine_readable_output)
	  {
	    Write_Header (m, r, params);
	    if (MC_iterations > 0)
	      {
		if (n_photons >= 0)
		  fprintf (stdout,
			   "#  Photons used to estimate lost light =   %ld\n",
			   n_photons);
		else
		  fprintf (stdout,
			   "#     Time used to estimate lost light =   %ld ms\n",
			   -n_photons);
	      }
	    else
	      fprintf (stdout,
		       "#  No photons were killed trying to figure out much light was lost\n");

	    fprintf (stdout, "#\n");
	    if (show_all)
	      {
		fprintf (stdout, "#      \tMeas\tCalc \t");
		fprintf (stdout, "Meas \tCalc \tBest \tBest \tBest \t");
		fprintf (stdout,
			 "Lost \tLost \tLost \tLost \tNone \tSpOnly\tMConly\t");
		fprintf (stdout, "None \tSpOnly\tMConly\tMC\tError\n");
		fprintf (stdout, "##wave\tM_R \tM_R  \t");
		fprintf (stdout,
			 "M_T \tM_T  \tmu_a\tmu_s'\t  g\tR_drct\tR_diff\tT_drct\tT_diff\t");
		fprintf (stdout,
			 "mu_a\tmu_a\tmu_a\tmu_s'\tmu_s'\tmu_s'\titer\tNumber\n");
		fprintf (stdout, "# [nm] \t[---]\t[---]\t[---]\t[---]\t");
		fprintf (stdout,
			 "[1/mm]\t[1/mm]\t[---]\t[---]\t[---]\t[---]\t");
		fprintf (stdout, "[1/mm]\t[1/mm]\t");
		fprintf (stdout,
			 "[1/mm]\t[1/mm]\t[1/mm]\t[1/mm]\t[---]\t[---]\n");
	      }
	    else
	      {
		fprintf (stdout, "#      \tMeas\tCalc\t");
		fprintf (stdout, "Meas\tCalc\t");
		fprintf (stdout, "Calc\tCalc\tCalc\tError\n");
		fprintf (stdout, "##wave\tM_R \tM_R \tM_T \tM_T \t");
		fprintf (stdout, "mu_a\tmu_s'\tg\tNumber\n");
		fprintf (stdout, "# [nm] \t[---]\t[---]\t[---]\t[---]\t");
		fprintf (stdout, "[1/mm]\t[1/mm]\t[---]\t[---]\n");
	      }
	  }

	m.ur1_lost = 0;
	m.ut1_lost = 0;
	m.uru_lost = 0;
	m.utu_lost = 0;
	r.error = IAD_NO_ERROR;
	m.f_r = 0.0;

	memcpy (&m_none, &m, sizeof (struct measure_type));
	memcpy (&r_none, &r, sizeof (struct invert_type));
	memcpy (&m_mc, &m, sizeof (struct measure_type));
	memcpy (&r_mc, &r, sizeof (struct invert_type));

	Inverse_RT (m, &r);
	calculate_coefficients (m, r, &LR, &LT, &mu_sp_sphere, &mu_a_sphere);


	mu_sp_both = mu_sp_sphere;
	mu_sp_none = mu_sp_both;
	mu_sp_mc = mu_sp_both;
	mu_a_both = mu_a_sphere;
	mu_a_mc = mu_a_both;
	mu_a_none = mu_a_both;


	if (m.num_spheres > 0 && r.error == IAD_NO_ERROR)
	  {

	    if (show_all)
	      {
		m_none.num_spheres = 0;
		Inverse_RT (m_none, &r_none);
		calculate_coefficients (m_none, r_none, &LR, &LT, &mu_sp_none,
					&mu_a_none);
	      }

	    while (mc_iter < MC_iterations && r.error == IAD_NO_ERROR)
	      {
		double ur1_max_loss, ut1_max_loss;
		double mu_sp_last = mu_sp_both;
		double mu_a_last = mu_a_both;

		MC_Lost (m, r, n_photons, &ur1, &ut1, &uru, &utu,
			 &ur1_lost, &ut1_lost, &uru_lost, &utu_lost);
		mc_iter_count++;
		mc_iter++;


		m.ur1_lost = ur1_lost;
		m.ut1_lost = ut1_lost;
		m.uru_lost = uru_lost;
		m.utu_lost = utu_lost;

		Max_Light_Loss (m, r, &ur1_max_loss, &ut1_max_loss);
		m.ur1_lost = ur1_max_loss;
		m.ut1_lost = ut1_max_loss;

		if (0 && ur1_lost + m.m_r + ut1_lost + m.m_t > 1)
		  {
		    struct AD_slab_type s;
		    double ad_ur1, ad_ut1, ad_uru, ad_utu;

		    s.a = r.a;
		    s.b = r.b;
		    s.g = r.g;
		    s.phase_function = HENYEY_GREENSTEIN;
		    s.n_slab = m.slab_index;
		    s.n_top_slide = m.slab_top_slide_index;
		    s.n_bottom_slide = m.slab_top_slide_index;
		    s.b_top_slide = 0;
		    s.b_bottom_slide = 0;

		    RT (32, &s, &ad_ur1, &ad_ut1, &ad_uru, &ad_utu);
		    fprintf (stderr, "UR1_AD=%7.5f UR1_MC=%7.5f ", ad_ur1,
			     ur1);
		    fprintf (stderr, "UR1_LOST=%7.5f UR1_V1=%7.5f\n",
			     ur1_lost, m.m_r);
		    fprintf (stderr, "UT1_AD=%7.5f UT1_MC=%7.5f ", ad_ut1,
			     ut1);
		    fprintf (stderr, "UT1_LOST=%7.5f UT1_V2=%7.5f\n",
			     ut1_lost, m.m_t);
		    r.error = IAD_EXCESSIVE_LIGHT_LOSS;
		    break;
		  }

		Inverse_RT (m, &r);
		calculate_coefficients (m, r, &LR, &LT, &mu_sp_both,
					&mu_a_both);


		if (show_all)
		  {
		    m_mc.num_spheres = 0;
		    m_mc.ur1_lost = ur1_lost;
		    m_mc.ut1_lost = ut1_lost;
		    m_mc.uru_lost = uru_lost;
		    m_mc.utu_lost = utu_lost;
		    Inverse_RT (m_mc, &r_mc);
		    calculate_coefficients (m_mc, r_mc, &LR, &LT, &mu_sp_mc,
					    &mu_a_mc);
		  }

		if (fabs (mu_a_last - mu_a_both) / (mu_a_both + 0.0001) <
		    r.MC_tolerance
		    && fabs (mu_sp_last - mu_sp_both) / (mu_sp_both +
							 0.0001) <
		    r.MC_tolerance)
		  break;

		if (!quiet)
		  print_dot (start_time, r.error, mc_iter_count,
			     rt_data_count, mc_iter, &any_error);

	      }
	  }

	if (m.lambda != 0)
	  fprintf (stdout, "%6.1f", m.lambda);
	else
	  fprintf (stdout, "      ");

	fprintf (stdout, "\t");

	fprintf (stdout, "%6.4f\t%6.4f\t", m.m_r, LR);
	fprintf (stdout, "%6.4f\t%6.4f\t", m.m_t, LT);
	fprintf (stdout, "%6.4f\t", mu_a_both);
	fprintf (stdout, "%6.4f\t", mu_sp_both);
	fprintf (stdout, "%6.4f\t", r.g);

	if (show_all)
	  {
	    fprintf (stdout, "%6.4f\t%6.4f\t", m.ur1_lost, m.uru_lost);
	    fprintf (stdout, "%6.4f\t%6.4f\t", m.ut1_lost, m.utu_lost);
	    fprintf (stdout, "%6.4f\t", mu_a_none);
	    fprintf (stdout, "%6.4f\t", mu_a_mc);
	    fprintf (stdout, "%6.4f\t", mu_a_sphere);

	    fprintf (stdout, "%6.4f\t", mu_sp_none);
	    fprintf (stdout, "%6.4f\t", mu_sp_mc);
	    fprintf (stdout, "%6.4f\t", mu_sp_sphere);
	    fprintf (stdout, "%2d\t", mc_iter);
	  }

	fprintf (stdout, "%2d\n", r.error);
	fflush (stdout);

	if (!quiet)
	  print_dot (start_time, r.error, mc_iter_count, rt_data_count, 99,
		     &any_error);

	r.error = IAD_NO_ERROR;
      }


      return 0;
    }


  if (argc > 1)
    {
      fprintf (stderr, "Only a single file can be processed at a time\n");
      fprintf (stderr, "try 'apply iad file1 file2 ... fileN'\n");
      exit (1);
    }

  if (argc == 1 && strcmp (argv[0], "-") != 0)
    {
      int n;
      char *base_name, *rt_name;
      base_name = strdup (argv[0]);
      n = (int) (strlen (base_name) - strlen (".rxt"));

      if (n > 0 && strstr (base_name + n, ".rxt") != NULL)
	base_name[n] = '\0';

      rt_name = strdup_together (base_name, ".rxt");

      if (freopen (argv[0], "r", stdin) == NULL &&
	  freopen (rt_name, "r", stdin) == NULL)
	{
	  fprintf (stderr, "Could not open either '%s' or '%s'\n",
		   argv[0], rt_name);
	  exit (1);
	}

      if (g_out_name == NULL)
	g_out_name = strdup_together (base_name, ".txt");

      free (rt_name);
      free (base_name);
    }

  if (g_out_name != NULL)
    {
      if (freopen (g_out_name, "w", stdout) == NULL)
	{
	  fprintf (stderr, "Could not open file '%s' for output\n",
		   g_out_name);
	  exit (1);
	}
    }



  if (Read_Header (stdin, &m, &params) == 0)
    {
      start_time = clock ();
      while (Read_Data_Line (stdin, &m, params) == 0)
	{

	  {
	    static int rt_data_count = 0;
	    static int mc_iter_count = 0;
	    struct measure_type m_mc, m_none;
	    struct invert_type r_mc, r_none;
	    int mc_iter = 0;

	    double ur1 = 0;
	    double ut1 = 0;
	    double uru = 0;
	    double utu = 0;
	    double ur1_lost = 0;
	    double ut1_lost = 0;
	    double uru_lost = 0;
	    double utu_lost = 0;
	    double mu_a_none = 0;
	    double mu_a_sphere = 0;
	    double mu_a_both = 0;
	    double mu_a_mc = 0;
	    double mu_sp_none = 0;
	    double mu_sp_sphere = 0;
	    double mu_sp_both = 0;
	    double mu_sp_mc = 0;
	    double LR = 0;
	    double LT = 0;

	    rt_data_count++;

	    Initialize_Result (m, &r);
	    r.method.quad_pts = cl_quadrature_points;

	    if (cl_default_a != UNINITIALIZED)
	      {
		r.default_a = cl_default_a;
	      }

	    if (cl_default_b != UNINITIALIZED)
	      {
		r.default_b = cl_default_b;
	      }

	    if (cl_default_g != UNINITIALIZED)
	      {
		r.default_g = cl_default_g;
	      }

	    if (cl_tolerance != UNINITIALIZED)
	      {
		r.tolerance = cl_tolerance;
		r.MC_tolerance = cl_tolerance;
	      }
	    r.search = determine_search (m, r);

	    if (rt_data_count == 1 && !machine_readable_output)
	      {
		Write_Header (m, r, params);
		if (MC_iterations > 0)
		  {
		    if (n_photons >= 0)
		      fprintf (stdout,
			       "#  Photons used to estimate lost light =   %ld\n",
			       n_photons);
		    else
		      fprintf (stdout,
			       "#     Time used to estimate lost light =   %ld ms\n",
			       -n_photons);
		  }
		else
		  fprintf (stdout,
			   "#  No photons were killed trying to figure out much light was lost\n");

		fprintf (stdout, "#\n");
		if (show_all)
		  {
		    fprintf (stdout, "#      \tMeas\tCalc \t");
		    fprintf (stdout, "Meas \tCalc \tBest \tBest \tBest \t");
		    fprintf (stdout,
			     "Lost \tLost \tLost \tLost \tNone \tSpOnly\tMConly\t");
		    fprintf (stdout, "None \tSpOnly\tMConly\tMC\tError\n");
		    fprintf (stdout, "##wave\tM_R \tM_R  \t");
		    fprintf (stdout,
			     "M_T \tM_T  \tmu_a\tmu_s'\t  g\tR_drct\tR_diff\tT_drct\tT_diff\t");
		    fprintf (stdout,
			     "mu_a\tmu_a\tmu_a\tmu_s'\tmu_s'\tmu_s'\titer\tNumber\n");
		    fprintf (stdout, "# [nm] \t[---]\t[---]\t[---]\t[---]\t");
		    fprintf (stdout,
			     "[1/mm]\t[1/mm]\t[---]\t[---]\t[---]\t[---]\t");
		    fprintf (stdout, "[1/mm]\t[1/mm]\t");
		    fprintf (stdout,
			     "[1/mm]\t[1/mm]\t[1/mm]\t[1/mm]\t[---]\t[---]\n");
		  }
		else
		  {
		    fprintf (stdout, "#      \tMeas\tCalc\t");
		    fprintf (stdout, "Meas\tCalc\t");
		    fprintf (stdout, "Calc\tCalc\tCalc\tError\n");
		    fprintf (stdout, "##wave\tM_R \tM_R \tM_T \tM_T \t");
		    fprintf (stdout, "mu_a\tmu_s'\tg\tNumber\n");
		    fprintf (stdout, "# [nm] \t[---]\t[---]\t[---]\t[---]\t");
		    fprintf (stdout, "[1/mm]\t[1/mm]\t[---]\t[---]\n");
		  }
	      }

	    m.ur1_lost = 0;
	    m.ut1_lost = 0;
	    m.uru_lost = 0;
	    m.utu_lost = 0;
	    r.error = IAD_NO_ERROR;
	    m.f_r = 0.0;

	    memcpy (&m_none, &m, sizeof (struct measure_type));
	    memcpy (&r_none, &r, sizeof (struct invert_type));
	    memcpy (&m_mc, &m, sizeof (struct measure_type));
	    memcpy (&r_mc, &r, sizeof (struct invert_type));

	    Inverse_RT (m, &r);
	    calculate_coefficients (m, r, &LR, &LT, &mu_sp_sphere,
				    &mu_a_sphere);


	    mu_sp_both = mu_sp_sphere;
	    mu_sp_none = mu_sp_both;
	    mu_sp_mc = mu_sp_both;
	    mu_a_both = mu_a_sphere;
	    mu_a_mc = mu_a_both;
	    mu_a_none = mu_a_both;


	    if (m.num_spheres > 0 && r.error == IAD_NO_ERROR)
	      {

		if (show_all)
		  {
		    m_none.num_spheres = 0;
		    Inverse_RT (m_none, &r_none);
		    calculate_coefficients (m_none, r_none, &LR, &LT,
					    &mu_sp_none, &mu_a_none);
		  }

		while (mc_iter < MC_iterations && r.error == IAD_NO_ERROR)
		  {
		    double ur1_max_loss, ut1_max_loss;
		    double mu_sp_last = mu_sp_both;
		    double mu_a_last = mu_a_both;

		    MC_Lost (m, r, n_photons, &ur1, &ut1, &uru, &utu,
			     &ur1_lost, &ut1_lost, &uru_lost, &utu_lost);
		    mc_iter_count++;
		    mc_iter++;


		    m.ur1_lost = ur1_lost;
		    m.ut1_lost = ut1_lost;
		    m.uru_lost = uru_lost;
		    m.utu_lost = utu_lost;

		    Max_Light_Loss (m, r, &ur1_max_loss, &ut1_max_loss);
		    m.ur1_lost = ur1_max_loss;
		    m.ut1_lost = ut1_max_loss;

		    if (0 && ur1_lost + m.m_r + ut1_lost + m.m_t > 1)
		      {
			struct AD_slab_type s;
			double ad_ur1, ad_ut1, ad_uru, ad_utu;

			s.a = r.a;
			s.b = r.b;
			s.g = r.g;
			s.phase_function = HENYEY_GREENSTEIN;
			s.n_slab = m.slab_index;
			s.n_top_slide = m.slab_top_slide_index;
			s.n_bottom_slide = m.slab_top_slide_index;
			s.b_top_slide = 0;
			s.b_bottom_slide = 0;

			RT (32, &s, &ad_ur1, &ad_ut1, &ad_uru, &ad_utu);
			fprintf (stderr, "UR1_AD=%7.5f UR1_MC=%7.5f ", ad_ur1,
				 ur1);
			fprintf (stderr, "UR1_LOST=%7.5f UR1_V1=%7.5f\n",
				 ur1_lost, m.m_r);
			fprintf (stderr, "UT1_AD=%7.5f UT1_MC=%7.5f ", ad_ut1,
				 ut1);
			fprintf (stderr, "UT1_LOST=%7.5f UT1_V2=%7.5f\n",
				 ut1_lost, m.m_t);
			r.error = IAD_EXCESSIVE_LIGHT_LOSS;
			break;
		      }

		    Inverse_RT (m, &r);
		    calculate_coefficients (m, r, &LR, &LT, &mu_sp_both,
					    &mu_a_both);


		    if (show_all)
		      {
			m_mc.num_spheres = 0;
			m_mc.ur1_lost = ur1_lost;
			m_mc.ut1_lost = ut1_lost;
			m_mc.uru_lost = uru_lost;
			m_mc.utu_lost = utu_lost;
			Inverse_RT (m_mc, &r_mc);
			calculate_coefficients (m_mc, r_mc, &LR, &LT,
						&mu_sp_mc, &mu_a_mc);
		      }

		    if (fabs (mu_a_last - mu_a_both) / (mu_a_both + 0.0001) <
			r.MC_tolerance
			&& fabs (mu_sp_last - mu_sp_both) / (mu_sp_both +
							     0.0001) <
			r.MC_tolerance)
		      break;

		    if (!quiet)
		      print_dot (start_time, r.error, mc_iter_count,
				 rt_data_count, mc_iter, &any_error);

		  }
	      }

	    if (m.lambda != 0)
	      fprintf (stdout, "%6.1f", m.lambda);
	    else
	      fprintf (stdout, "      ");

	    fprintf (stdout, "\t");

	    fprintf (stdout, "%6.4f\t%6.4f\t", m.m_r, LR);
	    fprintf (stdout, "%6.4f\t%6.4f\t", m.m_t, LT);
	    fprintf (stdout, "%6.4f\t", mu_a_both);
	    fprintf (stdout, "%6.4f\t", mu_sp_both);
	    fprintf (stdout, "%6.4f\t", r.g);

	    if (show_all)
	      {
		fprintf (stdout, "%6.4f\t%6.4f\t", m.ur1_lost, m.uru_lost);
		fprintf (stdout, "%6.4f\t%6.4f\t", m.ut1_lost, m.utu_lost);
		fprintf (stdout, "%6.4f\t", mu_a_none);
		fprintf (stdout, "%6.4f\t", mu_a_mc);
		fprintf (stdout, "%6.4f\t", mu_a_sphere);

		fprintf (stdout, "%6.4f\t", mu_sp_none);
		fprintf (stdout, "%6.4f\t", mu_sp_mc);
		fprintf (stdout, "%6.4f\t", mu_sp_sphere);
		fprintf (stdout, "%2d\t", mc_iter);
	      }

	    fprintf (stdout, "%2d\n", r.error);
	    fflush (stdout);

	    if (!quiet)
	      print_dot (start_time, r.error, mc_iter_count, rt_data_count,
			 99, &any_error);

	    r.error = IAD_NO_ERROR;
	  }


	  first_line = 0;
	}
    }
  if (!quiet)
    fprintf (stderr, "\n");
  if (any_error)
    print_error_legend ();
  return 0;
}
